﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;

public class WaterBehaviour : MonoBehaviour {

    [SerializeField]
    private float waterHeight;
    public float deltaWaterHeight;
    public float waterRemainder; //how much water remains in the tile, aka 1-this is what's being devided on the surrounding tiles

    public int x, y;

    // External variables
    public bool isOccupied;

    public bool flowing = true;

    public string occupantType; //Blockade, Divider, Drainage

    public GameObject occupier;

    // Is set from WaterIndex.cs
    public WaterIndex waterIndex;

    //public int direction = 0;
    public float[] direction; //0 = right, 1 = righttop, 2 = top etc. - 7 = rightbottom
    public bool hasDirection;

    private Gridfield groundTiles;

    private GameObject gameManager;

    private void Start()
    {
        direction = new float[8];
        for (int i = 0; i < direction.Length; i++)
        {
            direction[i] = (i + 1) % 2 + 1;
        }
    }

    private void LateUpdate()
    {
        WaterHeight += deltaWaterHeight;
        float drainRate = transform.parent.GetComponent<WaterIndex>().drainRate;
        if (waterHeight > drainRate)
            WaterHeight -= drainRate;
        else
            WaterHeight = 0;

        deltaWaterHeight = 0;
    }

    /// <summary>
    /// Update water movement
    /// </summary>
    public void UpdateWater()
    {
        
        if (WaterHeight > 0.01f && flowing)
        {
            GameObject[] neighbours = GetNeighbours();
            float stepWaterHeight = waterHeight;
            NewDirectionalFlow(stepWaterHeight);
        }

        //Make watertiles inside objects appear colloured somewhat      
        if (isOccupied && !occupier.GetComponent<PlacementProperties>().destroyed)
        {
            GameObject[] neighbours = GetNeighbours();
            Color color = GetComponent<SpriteRenderer>().color;
            float a = 0;
            for (int i = 0; i < 8; i += 2)
            {
                if (neighbours[i] != null)
                {
                    a += 1.0f - 1.0f / (neighbours[i].GetComponent<WaterBehaviour>().waterHeight + 1);
                }
            }
            a /= 2.0f;
            color.a = a;
            GetComponent<SpriteRenderer>().color = color;
        }
    }
    /// <summary>
    /// All the behaviour to see what neighbouring water tiles should get what amount of water, and all placeable objects' interaction with water
    /// </summary>
    /// <param name="stepWaterHeight"></param>
    private void NewDirectionalFlow(float stepWaterHeight)
    {
        GameObject[] neighbours = GetNeighbours();
        float totalFraction = GetTotalFraction();
        float drainFactor = 1;
        

        //Check all neighbours
        for (int i = 0; i < neighbours.Length; i++)
        {
            //if it's the edge of the map, reflect
            if (neighbours[i] == null)
            {
                direction[i] = 0;
                totalFraction = GetTotalFraction();
            }
            else {
                WaterBehaviour neighbourScript = neighbours[i].GetComponent<WaterBehaviour>();
                if (neighbourScript.isOccupied)
                {
                    PlacementProperties occupierScript = neighbourScript.occupier.GetComponent<PlacementProperties>();
                    //if it's blockade, check for the height, overflow it if nessecairy, take away health
                    if (neighbourScript.occupantType == "Blockade")
                    {
                        if (!occupierScript.destroyed && occupierScript.strength > waterHeight * (1 - waterRemainder) * direction[i] / totalFraction)
                        {
                            occupierScript.health -= waterHeight * (1 - waterRemainder) * direction[i] / totalFraction;
                            if (occupierScript.health < 0)
                            {
                                occupierScript.DestroyPlacement();
                            }
                            direction[i] = 0;
                            totalFraction = GetTotalFraction();
                        }
                        //overflow over the blockacde
                        else if (!occupierScript.destroyed)
                        {
                            occupierScript.health -= waterHeight * (1 - waterRemainder) * direction[i] / totalFraction;
                            if (occupierScript.health < 0)
                            {
                                occupierScript.DestroyPlacement();
                            }
                            direction[i] *= 1 - occupierScript.strength / (waterHeight * (1 - waterRemainder) * direction[i] / totalFraction);
                            totalFraction = GetTotalFraction();
                        }
                    }
                    //If it's a divider, drain some of the water, and deflect it partially
                    else if (neighbourScript.occupantType == "Divider")
                    {
                        occupierScript.health -= waterHeight * (1 - waterRemainder) * direction[i] / totalFraction;
                        drainFactor = 0.8f;
                        if (occupierScript.health < 0)
                        {
                            occupierScript.DestroyPlacement();
                        }
                        direction[i] /= occupierScript.strength;
                        totalFraction = GetTotalFraction();
                    }
                    //If it's a drain, just remove most of the water.
                    else if (neighbourScript.occupantType == "Drainage")
                    {
                        drainFactor = 1 / occupierScript.strength;
                    }
                }
            }
        }

        // Flow to neighbours
        for (int i = 0; i < neighbours.Length; i++)
        {
            if (neighbours[i] != null)
            {
                neighbours[i].GetComponent<WaterBehaviour>().deltaWaterHeight += (1 - waterRemainder) * (direction[i] / totalFraction) * stepWaterHeight * drainFactor;
                neighbours[i].GetComponent<WaterBehaviour>().direction[i] = direction[i]; //for example, the neighbour right of this one, will get this tile's right movement shoved into its direction set.
                deltaWaterHeight -= (1 - waterRemainder) * (direction[i] / totalFraction) * stepWaterHeight;
               
            }
        }
    }
/// <summary>
/// Calculates the total amount of dirction, needed to get the actual fraction of the total each direction has
/// </summary>
/// <returns></returns>
   private float GetTotalFraction()
    {
        float totalFraction = 0;
        for (int i = 0; i < direction.Length; i++)
        {
            totalFraction += direction[i];
        }
        //if there's no direction anywhere, redo the standard direction.
        if (totalFraction == 0)
        {
            direction = new float[8];
            for (int i = 0; i < direction.Length; i++)
            {
                direction[i] = (i + 1) % 2 + 1;
            }
            totalFraction = 12;
        }
       
        return totalFraction;
    }


    /// <summary>
    /// Returns water tiles
    /// </summary>
    /// <returns>Water neighbour tiles, counterclockwise</returns>
    private GameObject[] GetNeighbours()
    {
        GameObject[] neighbours = new GameObject[8];

        // Right neigbour
        neighbours[0] = waterIndex.GetTile(x + 1, y);

        // Upper right
        neighbours[1] = waterIndex.GetTile(x + 1, y + 1);

        // Upper neighbour
        neighbours[2] = waterIndex.GetTile(x, y + 1);

        // Upper left
        neighbours[3] = waterIndex.GetTile(x - 1, y + 1);

        // Left neighbour
        neighbours[4] = waterIndex.GetTile(x - 1, y);

        // Lower left
        neighbours[5] = waterIndex.GetTile(x - 1, y - 1);

        // Lower neighbour
        neighbours[6] = waterIndex.GetTile(x, y - 1);

        // Lower right
        neighbours[7] = waterIndex.GetTile(x + 1, y - 1);

        return neighbours;
    }
    
    public float WaterHeight
    {
        get { return waterHeight; }
        set
        {
            waterHeight = value;

            // Change opacity if the tile is not occupied
            if (!isOccupied || occupier.GetComponent<PlacementProperties>().destroyed)
            {
                Color color = GetComponent<SpriteRenderer>().color;
                color.a = 1.0f - 1.0f / (waterHeight + 1);

                GetComponent<SpriteRenderer>().color = color;
            }
        }
    }
}
